"""
Methods to provide code coverage using coverage.py.
"""

import os
import sys
import webbrowser

try:
    from coverage import coverage
except ImportError:
    coverage = None


# use to hold a global coverage obj
_coverobj = None

def setup_coverage(options):
    global _coverobj
    if _coverobj is None and (options.coverage or options.coveragehtml):
        if not coverage:
            raise RuntimeError("coverage has not been installed.")
        if not options.coverpkgs:
            raise RuntimeError("No packages specified for coverage. "
                               "Use the --coverpkg option to add a package.")
        _coverobj = coverage(data_suffix=True, source=options.coverpkgs,
                             omit=options.cover_omits)
    return _coverobj

def start_coverage():
    if _coverobj:
        _coverobj.start()

def stop_coverage():
    if _coverobj:
        _coverobj.stop()

def save_coverage():
    if _coverobj:
        _coverobj.save()

def finalize_coverage(options):
    if _coverobj and options.coverpkgs:
        rank = 0
        if not options.nompi:
            try:
                from mpi4py import MPI
                rank = MPI.COMM_WORLD.rank
            except ImportError:
                pass
        if rank == 0:
            from testflo.util import find_files, find_module
            excl = lambda n: (n.startswith('test_') and n.endswith('.py')) or \
                             n.startswith('__init__.')
            dirs = []
            for n in options.coverpkgs:
                if os.path.isdir(n):
                    dirs.append(n)
                else:
                    path = find_module(n)
                    if path is None:
                        raise RuntimeError("Can't find module %s" % n)
                    dirs.append(os.path.dirname(path))

            morfs = list(find_files(dirs, match='*.py', exclude=excl))

            _coverobj.combine()

            # write combined data to default filename (as used by coveralls)
            # (NOTE: get_data() returns None, so using data attribute)
            _coverobj.data.write_file('.coverage')

            if options.coverage:
                _coverobj.report(morfs=morfs)
            else:
                dname = '_html'
                _coverobj.html_report(morfs=morfs, directory=dname)
                outfile = os.path.join(os.getcwd(), dname, 'index.html')

                if sys.platform == 'darwin':
                    os.system('open %s' % outfile)
                else:
                    webbrowser.get().open(outfile)



from __future__ import print_function

class TimeFilter(object):
    """This iterator saves to the specified output file only those tests
    that complete successfully in max_time seconds or less.  This output
    file can later be fed into testflo to limit the tests to those in the
    file.
    """
    def __init__(self, max_time, outfile='quicktests.in'):
        self.outfile = outfile
        self.max_time = max_time

    def get_iter(self, input_iter):
        with open(self.outfile, 'w') as f:
            for result in input_iter:
                if result.status == 'OK' and result.elapsed() <= self.max_time:
                    print(result.spec, file=f)
                yield result


class FailFilter(object):
    """This iterator saves to the specified output file only those tests
    that fail.  This output file can later be fed into testflo to limit the
    tests to those in the file.
    """
    def __init__(self, outfile='failtests.in'):
        self.outfile = outfile

    def get_iter(self, input_iter):
        with open(self.outfile, 'w') as f:
            for result in input_iter:
                if result.status == 'FAIL' and not result.expected_fail:
                    print(result.spec, file=f)
                yield result



import sys
import os

class DurationSummary(object):
    """Writes a summary of the tests taking the longest time."""

    def __init__(self, options, stream=sys.stdout):
        self.stream = stream
        self.options = options
        self.startdir = os.getcwd()

    def get_iter(self, input_iter):
        durations = []

        for test in input_iter:
            durations.append((test.spec, test.end_time - test.start_time))
            yield test

        write = self.stream.write
        mintime = self.options.durations_min

        if mintime > 0.:
            title = " Max duration tests with duration >= {} sec ".format(mintime)
        else:
            title = " Max duration tests "

        eqs = "=" * 16

        write("\n\n{}{}{}\n\n".format(eqs, title, eqs))
        count = self.options.durations

        for spec, duration in sorted(durations, key=lambda t: t[1], reverse=True):
            if duration < mintime:
                break

            if spec.startswith(self.startdir):
                spec = spec[len(self.startdir):]

            write("{:8.3f} sec - {}\n".format(duration, spec))

            count -= 1
            if count <= 0:
                break

        write("\n" + "=" * (len(title) + 2 * len(eqs)) + "\n")



"""
Methods and class for running tests.
"""
from __future__ import print_function

import sys
import os

from multiprocessing import Queue, Process

from testflo.cover import save_coverage
from testflo.test import Test
from testflo.options import get_options


def worker(test_queue, done_queue, subproc_queue, worker_id):
    """This is used by concurrent test processes. It takes a test
    off of the test_queue, runs it, then puts the Test object
    on the done_queue.
    """
    test_count = 0
    for tests in iter(test_queue.get, 'STOP'):

        done_tests = []
        for test in tests:
            try:
                test_count += 1
                done_tests.append(test.run(subproc_queue))
            except:
                # we generally shouldn't get here, but just in case,
                # handle it so that the main process doesn't hang at the
                # end when it tries to join all of the concurrent processes.
                done_tests.append(test)

        done_queue.put(done_tests)

    # don't save anything unless we actually ran a test
    if test_count > 0:
        save_coverage()


class TestRunner(object):

    def __init__(self, options, subproc_queue):
        self.stop = options.stop
        self.pre_announce = options.pre_announce
        self._queue = subproc_queue

    def get_iter(self, input_iter):
        """Run tests serially."""

        for tests in input_iter:
            stop = False
            for test in tests:
                if self.pre_announce:
                    print("    about to run %s " % test.short_name(), end='')
                    sys.stdout.flush()
                result = test.run(self._queue)
                yield result
                if self.stop:
                    if (result.status == 'FAIL' and not result.expected_fail) or (
                                  result.status == 'OK' and result.expected_fail):
                          stop = True
                          break
            if stop:
                break

        save_coverage()


class ConcurrentTestRunner(TestRunner):
    """TestRunner that uses the multiprocessing package
    to execute tests concurrently.
    """

    def __init__(self, options, subproc_queue):
        super(ConcurrentTestRunner, self).__init__(options, subproc_queue)
        self.num_procs = options.num_procs

        # only do concurrent stuff if num_procs > 1
        if self.num_procs > 1:
            self.get_iter = self.run_concurrent_tests

            # Create queues
            self.task_queue = Queue()
            self.done_queue = Queue()

            self.procs = []

            # Start worker processes
            for i in range(self.num_procs):
                worker_id = "%d_%d" % (os.getpid(), i)
                self.procs.append(Process(target=worker,
                                          args=(self.task_queue,
                                                self.done_queue, subproc_queue,
                                                worker_id)))

            for proc in self.procs:
                proc.start()

    def run_concurrent_tests(self, input_iter):
        """Run tests concurrently."""

        it = iter(input_iter)
        numtests = 0
        try:
            for proc in self.procs:
                self.task_queue.put(next(it))
                numtests += 1
        except StopIteration:
            pass
        else:
            try:
                while numtests:
                    stop = False
                    results = self.done_queue.get()
                    for result in results:
                        yield result
                        if self.stop:
                            if (result.status == 'FAIL' and not result.expected_fail) or (
                                  result.status == 'OK' and result.expected_fail):
                                stop = True
                                break
                    if stop:
                        break
                    numtests -= 1
                    self.task_queue.put(next(it))
                    numtests += 1
            except StopIteration:
                pass

        for proc in self.procs:
            self.task_queue.put('STOP')

        for i in range(numtests):
            results = self.done_queue.get()
            for result in results:
                yield result

        for proc in self.procs:
            proc.join()




"""
This is meant to be executed using mpirun.  It is called as a subprocess
to run an MPI test.

"""

if __name__ == '__main__':
    import sys
    import os
    import traceback

    os.environ['OPENMDAO_USE_MPI'] = '1'

    from mpi4py import MPI
    from testflo.test import Test
    from testflo.cover import setup_coverage, save_coverage
    from testflo.qman import get_client_queue
    from testflo.options import get_options

    exitcode = 0  # use 0 for exit code of all ranks != 0 because otherwise,
                  # MPI will terminate other processes

    queue = get_client_queue()
    os.environ['TESTFLO_QUEUE'] = ''

    options = get_options()
    setup_coverage(options)

    try:
        try:
            comm = MPI.COMM_WORLD
            test = Test(sys.argv[1], options)
            test.nocapture = True # so we don't lose stdout
            test.run()
        except:
            print(traceback.format_exc())
            test.status = 'FAIL'
            test.err_msg = traceback.format_exc()

        # collect results
        results = comm.gather(test, root=0)
        if comm.rank == 0:
            if not all([isinstance(r, Test) for r in results]):
                print("\nNot all results gathered are Test objects.  "
                      "You may have out-of-sync collective MPI calls.\n")
            total_mem_usage = sum(r.memory_usage for r in results if isinstance(r, Test))
            test.memory_usage = total_mem_usage

            # check for errors and record error message
            for r in results:
                if test.status != 'FAIL' and r.status in ('SKIP', 'FAIL'):
                    test.err_msg = r.err_msg
                    test.status = r.status
                    if r.status == 'FAIL':
                        break

        save_coverage()

    except Exception:
        test.err_msg = traceback.format_exc()
        test.status = 'FAIL'

    finally:
        sys.stdout.flush()
        sys.stderr.flush()

        if comm.rank == 0:
            queue.put(test)



"""
Misc. utility routines.
"""

import os
import sys
import itertools
import inspect
import importlib
import warnings
from importlib import import_module

from configparser import ConfigParser

from fnmatch import fnmatch
from os.path import join, dirname, basename, isfile,  abspath, split, splitext

from argparse import ArgumentParser, _AppendAction

from testflo.cover import start_coverage, stop_coverage

_store = {}


def _get_parser():
    """Returns a parser to handle command line args."""

    parser = ArgumentParser()
    parser.usage = "testflo [options]"
    parser.add_argument('--version', action='store_true', dest='version',
                        help="Display the version number and exit.")
    parser.add_argument('-c', '--config', action='store', dest='cfg',
                        metavar='FILE',
                        help='Path of config file where preferences are specified.')
    parser.add_argument('-t', '--testfile', action='store', dest='testfile',
                        metavar='FILE',
                        help='Path to a file containing one testspec per line.')
    parser.add_argument('--maxtime', action='store', dest='maxtime',
                        metavar='TIME_LIMIT', default=-1, type=float,
                        help='Specifies a time limit in seconds for tests to be saved to '
                             'the quicktests.in file.')

    parser.add_argument('-n', '--numprocs', type=int, action='store',
                        dest='num_procs', metavar='NUM_TEST_PROCS',
                        help='Number of concurrent test processes to run. By default, this will '
                             'use the number of virtual processors available.  To force tests to '
                             'run consecutively, specify a value of 1.')
    parser.add_argument('-o', '--outfile', action='store', dest='outfile',
                        metavar='FILE', default='testflo_report.out',
                        help='Name of test report file.  Default is testflo_report.out.')
    parser.add_argument('-v', '--verbose', action='store_true', dest='verbose',
                        help="Include testspec and elapsed time in "
                             "screen output. Also shows all stderr output, even if test doesn't fail")
    parser.add_argument('--compact', action='store_true', dest='compact',
                        help="Limit output to a single character for each test.")
    parser.add_argument('--dryrun', action='store_true', dest='dryrun',
                        help="Don't actually run tests, but print "
                          "which tests would have been run.")
    parser.add_argument('--pre_announce', action='store_true', dest='pre_announce',
                        help="Announce the name of each test before it runs. This "
                             "can help track down a hanging test. This automatically sets -n 1.")
    parser.add_argument('-f', '--fail', action='store_true', dest='save_fails',
                        help="Save failed tests to failtests.in file.")
    parser.add_argument('--full_path', action='store_true', dest='full_path',
                        help="Display full test specs instead of shortened names.")
    parser.add_argument('-i', '--isolated', action='store_true', dest='isolated',
                        help="Run each test in a separate subprocess.")
    parser.add_argument('--nompi', action='store_true', dest='nompi',
                        help="Force all tests to run without MPI. This can be useful "
                             "for debugging.")
    parser.add_argument('-x', '--stop', action='store_true', dest='stop',
                        help="Stop after the first test failure, or as soon as possible"
                             " when running concurrent tests.")
    parser.add_argument('-s', '--nocapture', action='store_true', dest='nocapture',
                        help="Standard output (stdout) will not be captured and will be"
                             " written to the screen immediately.")
    parser.add_argument('--coverage', action='store_true', dest='coverage',
                        help="Perform coverage analysis and display results on stdout")
    parser.add_argument('--coverage-html', action='store_true', dest='coveragehtml',
                        help="Perform coverage analysis and display results in browser")
    parser.add_argument('--coverpkg', action='append', dest='coverpkgs',
                        metavar='PKG',
                        help="Add the given package to the coverage list. You"
                              " can use this option multiple times to cover"
                              " multiple packages.")
    parser.add_argument('--cover-omit', action='append', dest='cover_omits',
                        metavar='FILE',
                        help="Add a file name pattern to remove it from coverage.")

    parser.add_argument('-b', '--benchmark', action='store_true', dest='benchmark',
                        help='Specifies that benchmarks are to be run rather '
                             'than tests, so only files starting with "benchmark_" '
                             'will be executed.')
    parser.add_argument('-d', '--datafile', action='store', dest='benchmarkfile',
                        metavar='FILE', default='benchmark_data.csv',
                        help='Name of benchmark data file.  Default is benchmark_data.csv.')

    parser.add_argument('--durations', action='store', type=int, dest='durations', default=0,
                        metavar='NUM',
                        help="Display 'NUM' tests with longest durations.")

    parser.add_argument('--durations-min', action='store', type=float, dest='durations_min',
                        default=0.005, metavar='MIN_TIME',
                        help='Specify the minimum duration test to include in the durations list.')

    parser.add_argument('--noreport', action='store_true', dest='noreport',
                        help="Don't create a test results file.")

    parser.add_argument('--show_skipped', action='store_true', dest='show_skipped',
                        help="Display a list of any skipped tests in the summary.")

    parser.add_argument('--disallow_skipped', action='store_true', dest='disallow_skipped',
                        help="Return exit code 2 if no tests failed but some tests are skipped.")

    parser.add_argument('--show_deprecations', action='store_true', dest='show_deprecations',
                        help="Display a list of all deprecation warnings encountered in testing.")
    parser.add_argument('--deprecations_report', action='store', dest='deprecations_report',
                        metavar='FILE', default=None,
                        help='Generate a deprecations report with the given file name.  Default is None.')
    parser.add_argument('--disallow_deprecations', action='store_true', dest='disallow_deprecations',
                        help="Raise deprecation warnings as Exceptions.")

    parser.add_argument('tests', metavar='test', nargs='*',
                        help='A test method, test case, module, or directory to run. If not '
                             'supplied, the current working directory is assumed.')

    parser.add_argument('-m', '--match', '--testmatch', action='append', dest='test_glob',
                        metavar='GLOB',
                        help='Pattern to use for test discovery. Multiple patterns are allowed.',
                        default=[])

    parser.add_argument('--exclude', action='append', dest='excludes', metavar='GLOB', default=[],
                        help="Pattern to exclude test functions. Multiple patterns are allowed.")

    parser.add_argument('--skip_dir', action='append', dest='skip_dirs', metavar='GLOB', default=[],
                        help="Pattern to skip directories. Multiple patterns are allowed. Patterns "
                        "are applied only to local dir names, not full paths.")

    parser.add_argument('--timeout', action='store', dest='timeout', type=float, metavar='TIME_LIMIT',
                        help="Timeout in seconds. A test will be terminated if it takes longer than "
                             "'TIME_LIMIT'. Only works for tests running in a subprocess "
                             "(MPI or isolated).")

    return parser


def _options2args():
    """Gets the testflo args that should be used in subprocesses."""

    cmdset = set([
      '--nocapture',
      '-s',
      '--coverpkg',
      '--coverage',
      '--coverage-html',
      '--cover-omit',
    ])

    keep = []
    i = 0
    args = sys.argv[1:]
    argslen = len(args)
    while i < argslen:
        arg = args[i]
        if arg.split('=',1)[0] in cmdset:
            keep.append(arg)
            if ((arg.startswith('--coverpkg') or arg.startswith('--cover-omit'))
                        and '=' not in arg):
                i += 1
                keep.append(args[i])
        i += 1

    return keep


def _file_gen(dname, fmatch=bool, dmatch=None):
    """A generator returning files under the given directory, with optional
    file and directory filtering.

    fmatch: predicate funct
        A predicate function that returns True on a match.
        This is used to match files only.

    dmatch: predicate funct
        A predicate function that returns True on a match.
        This is used to match directories only.
    """
    if dmatch is not None and not dmatch(dname):
        return

    for path, dirlist, filelist in os.walk(dname):
        if dmatch is not None:  # prune directories to search
            newdl = [d for d in dirlist if dmatch(d)]
            if len(newdl) != len(dirlist):
                dirlist[:] = newdl  # replace contents of dirlist to cause pruning

        for name in [f for f in filelist if fmatch(f)]:
            yield join(path, name)


def find_files(start, match=None, exclude=None,
               dirmatch=None, direxclude=None):
    """Return filenames (using a generator).

    start: str or list of str
        Starting directory or list of directories.

    match: str or predicate funct
        Either a string containing a glob pattern to match
        or a predicate function that returns True on a match.
        This is used to match files only.

    exclude: str or predicate funct
        Either a string containing a glob pattern to exclude
        or a predicate function that returns True to exclude.
        This is used to exclude files only.

    dirmatch: str or predicate funct
        Either a string containing a glob pattern to match
        or a predicate function that returns True on a match.
        This is used to match directories only.

    direxclude: str or predicate funct
        Either a string containing a glob pattern to exclude
        or a predicate function that returns True to exclude.
        This is used to exclude directories only.

    Walks all subdirectories below each specified starting directory,
    subject to directory filtering.

    """
    startdirs = [start] if isinstance(start, str) else start
    if len(startdirs) == 0:
        return iter([])

    if match is None:
        matcher = bool
    elif isinstance(match, str):
        matcher = lambda name: fnmatch(name, match)
    else:
        matcher = match

    if dirmatch is None:
        dmatcher = bool
    elif isinstance(dirmatch, str):
        dmatcher = lambda name: fnmatch(name, dirmatch)
    else:
        dmatcher = dirmatch

    if isinstance(exclude, str):
        fmatch = lambda name: matcher(name) and not fnmatch(name, exclude)
    elif exclude is not None:
        fmatch = lambda name: matcher(name) and not exclude(name)
    else:
        fmatch = matcher

    if isinstance(direxclude, str):
        if dmatcher is bool:
            dmatch = lambda name: not fnmatch(name, direxclude)
        else:
            dmatch = lambda name: dmatcher(name) and not fnmatch(name, direxclude)
    elif direxclude is not None:
        dmatch = lambda name: dmatcher(name) and not direxclude(name)
    else:
        dmatch = dmatcher

    iters = [_file_gen(d, fmatch=fmatch, dmatch=dmatch) for d in startdirs]
    if len(iters) > 1:
        return itertools.chain(*iters)
    else:
        return iters[0]


def fpath2modpath(fpath):
    """Given a module filename, return its full Python name including
    enclosing packages. (based on existence of ``__init__.py`` files)
    """
    if basename(fpath).startswith('__init__.'):
        pnames = []
    else:
        pnames = [splitext(basename(fpath))[0]]

    path = dirname(abspath(fpath))

    while isfile(join(path, '__init__.py')):
        path, pname = split(path)
        pnames.append(pname)

    return '.'.join(pnames[::-1])


def parent_dirs(fpath):
    """Return a list of the absolute paths of the parent directory and
    each of its parent directories for the given file.
    """
    parts = abspath(fpath).split(os.path.sep)
    pdirs = []
    for i in range(2, len(parts)):
        pdirs.append(os.path.sep.join(parts[:i]))
    return pdirs[::-1]


def get_testpath(testspec):
    """Return the path to the test module separated from
    the rest of the test spec.
    """
    testspec = testspec.strip()
    parts = testspec.split(':')
    if len(parts) > 1 and parts[1].startswith('\\'):  # windows abs path
        path = ':'.join(parts[:2])
        if len(parts) == 3:
            rest = parts[2]
        else:
            rest = ''
    else:
        path, _, rest = testspec.partition(':')
    return path, rest


def find_module(name):
    """Return the pathname of the Python file corresponding to the
    given module name, or None if it can't be found. The
    file must be an uncompiled Python (.py) file.
    """
    try:
        info = importlib.util.find_spec(name)
    except ImportError:
        info = None
    if info is not None:
        return info.origin


_mod2file = {}  # keep track of non-pkg files to detect and flag dups


def try_import(fname, modpath):
    try:
        mod = import_module(modpath)
    except ImportError:
        # this might be a module that's not in the same
        # environment as testflo, so try temporarily prepending
        # its parent dirs to sys.path so it'll (hopefully) be
        # importable
        oldpath = sys.path[:]
        sys.path.extend(parent_dirs(fname))
        sys.path.append(os.getcwd())
        try:
            mod = import_module(modpath)
            # don't keep this module around in sys.modules, but
            # keep a reference to it, else multiprocessing on Windows
            # will have problems
            _store[modpath] = mod
            del sys.modules[modpath]
        finally:
            sys.path = oldpath

    return mod


def get_module(fname):
    """Given a filename or module path name, return a tuple
    of the form (filename, module).
    """

    if fname.endswith('.py'):
        modpath = fpath2modpath(fname)
        if not modpath:
            raise RuntimeError("can't find module %s" % fname)

        if modpath in _mod2file:
            old = _mod2file[modpath]
            if old != fname:
                raise RuntimeError("module '%s' was already imported earlier from file '%s' so "
                                    "it can't be imported from file '%s'. To fix this problem, "
                                    "either rename the file or add the file to a python package "
                                    "so the resulting module path will be unique." %
                                    (modpath, old, fname))
        else:
            _mod2file[modpath] = fname

    else:
        modpath = fname
        fname = find_module(modpath)

        if fname:
            _mod2file[modpath] = fname
        else:
            # check for a non-pkg module
            if modpath in _mod2file:
                fname = _mod2file[modpath]
            else:
                raise ImportError("can't import %s" % modpath)

    start_coverage()

    try:
        mod = try_import(fname, modpath)
    finally:
        stop_coverage()

    return fname, mod


def read_test_file(testfile):
    """Reads a file containing one testspec per line."""
    with open(os.path.abspath(testfile), 'r') as f:
        for line in f:
            idx = line.find('#')
            if idx >= 0:
                line = line[:idx]

            line = line.strip()
            if line:
                yield line


_parser_types = None


def _get_parser_action_map():
    global _parser_types

    if _parser_types is None:
        _parser_types = {}
        p = _get_parser()
        for action in p._actions:
            _parser_types[action.dest] = action

    return _parser_types


def read_config_file(cfgfile, options):
    config = ConfigParser()
    config.read_file(open(cfgfile), source=cfgfile)

    if 'testflo' in config:
        parser_map = _get_parser_action_map()

        for name, optstr in config['testflo'].items():
            if name not in parser_map:
                warnings.warn("Unknown option '{}' in testflo config file '{}'.".format(name,
                                                                                        cfgfile))
                continue

            action = parser_map[name]
            typ = action.type
            if typ is None:
                typ = lambda x: x

            if isinstance(action, _AppendAction):
                setattr(options, name, [typ(s.strip()) for s in optstr.split(',') if s.strip()])
            else:
                setattr(options, name, typ(optstr))


def get_memory_usage():
    """return memory usage for the current process"""
    k = 1024.
    try:
        # prefer psutil, it works on all platforms including Windows
        import psutil
        process = psutil.Process(os.getpid())
        mem = process.memory_info().rss
        return mem/(k*k)
    except ImportError:
        try:
            # fall back to getrusage, which works only on Linux and OSX
            import resource
            mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
            if sys.platform == 'darwin':
                return mem/(k*k)
            else:
                return mem/k
        except:
            return 0.


def elapsed_str(elapsed):
    """return a string of the form hh:mm:sec"""
    hrs = int(elapsed/3600)
    elapsed -= (hrs * 3600)
    mins = int(elapsed/60)
    elapsed -= (mins * 60)
    return "%02d:%02d:%.2f" % (hrs, mins, elapsed)


# in python3, inspect.ismethod doesn't work as you might expect, so...
def ismethod(obj):
    return inspect.isfunction(obj) or inspect.ismethod(obj)



"""
testflo is a python testing framework that takes an iterator of test
specifier names e.g., <test_module>:<testcase>.<test_method>, and feeds
them through a pipeline of iterators that operate on them and transform them
into Test objects, then pass them on to other objects in the pipeline.

The Discoverer object takes an initial list of directory names, module names,
or test specifier strings and returns an iterator of Test objects representing
all of the tests found.  The rest of the pipeline operates on these Test
objects, which contain a test specifier string, a status indicating
whether the test passed or failed, and captured stderr from the
running of the test.

The only real API function for objects added to the pipline is:

    def get_iter(self, input_iter)

Functions can also be added directly to the pipeline as long as they
take a Test object iterator as an arg and return a Test object iterator.

The get_iter function should expect to receive an iterator over Test objects
and should return an iterator over Test objects.

"""
from __future__ import print_function

import os
import sys
import time
import warnings
import multiprocessing

from fnmatch import fnmatch, fnmatchcase

import testflo
from testflo.runner import ConcurrentTestRunner
from testflo.printer import ResultPrinter
from testflo.benchmark import BenchmarkWriter
from testflo.summary import ResultSummary
from testflo.deprecations import DeprecationsReport
from testflo.duration import DurationSummary
from testflo.discover import TestDiscoverer
from testflo.filters import TimeFilter, FailFilter

from testflo.util import read_config_file, read_test_file, _get_parser
from testflo.cover import setup_coverage, finalize_coverage
from testflo.options import get_options
from testflo.qman import get_server_queue

options = get_options()


def dryrun(input_iter):
    """Iterator added to the pipeline when user only wants a dry run, listing specs for all of the
    discovered tests but not actually running them.
    """
    for tests in input_iter:
        for test in tests:
            if test.status is None:
                test.status = 'OK'
            print(test.spec)
            yield test


def run_pipeline(source, pipe, disallow_skipped):
    """Run a pipeline of test iteration objects."""

    global _start_time
    _start_time = time.time()

    iters = [source]

    # give each object the iterator from upstream in the pipeline
    for i,p in enumerate(pipe):
        iters.append(p(iters[i]))

    n_failed = 0
    n_skipped = 0
    # iterate over the last iter in the pipeline and we're done
    for result in iters[-1]:
        if result.status == 'FAIL' and not result.expected_fail:
            n_failed += 1
        elif result.status == 'SKIP':
            n_skipped += 1

    if n_failed > 0:
        return_code = 1
    elif n_skipped > 1 and disallow_skipped:
        return_code = 2
    else:
        return_code = 0

    return return_code


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    options = get_options(args)

    if options.version:
        print("testflo version %s" % testflo.__version__)
        return 0

    nprocs = options.num_procs

    options.skip_dirs = []

    # read user prefs from ~/.testflo file.
    # create one if it doesn't exist
    homedir = os.path.expanduser('~')
    rcfile = os.path.join(homedir, '.testflo')
    if not os.path.isfile(rcfile):
        with open(rcfile, 'w') as f:
            f.write("""[testflo]
skip_dirs=site-packages,
    dist-packages,
    build,
    _build,
    contrib
""")
    read_config_file(rcfile, options)
    if options.cfg:
        read_config_file(options.cfg, options)

    if nprocs is None and options.num_procs is None:
        try:
            options.num_procs = multiprocessing.cpu_count()
        except:
            warnings.warn('CPU count could not be determined. Defaulting to 1')
            options.num_procs = 1

    if nprocs is not None:
        options.num_procs = nprocs

    tests = options.tests
    if options.testfile:
        tests += list(read_test_file(options.testfile))

    if not tests:
        tests = [os.getcwd()]

    def dir_exclude(d):
        base = os.path.basename(d)
        for skip in options.skip_dirs:
            if fnmatch(base, skip):
                return True
        return False

    # set this so code will know when it's running under testflo
    os.environ['TESTFLO_RUNNING'] = '1'

    setup_coverage(options)

    if options.noreport:
        report_file = open(os.devnull, 'a')
    else:
        report_file = open(options.outfile, 'w')

    if not options.test_glob:
        options.test_glob = ['test*']

    def func_matcher(funcname):
        for pattern in options.excludes:
            if fnmatchcase(funcname, pattern):
                return False

        for pattern in options.test_glob:
            if fnmatchcase(funcname, pattern):
                return True

        return False

    if options.benchmark:
        options.num_procs = 1
        options.isolated = True
        discoverer = TestDiscoverer(options, module_pattern='benchmark*.py',
                                    func_match=lambda f: fnmatchcase(f, 'benchmark*'),
                                    dir_exclude=dir_exclude)
        benchmark_file = open(options.benchmarkfile, 'a')
    else:
        discoverer = TestDiscoverer(options, dir_exclude=dir_exclude,
                                    func_match=func_matcher)
        benchmark_file = open(os.devnull, 'a')

    retval = 0

    if options.isolated or not options.nompi:
        # create a distributed queue and get a proxy to it
        manager, queue = get_server_queue()
    else:
        manager, queue = (None, None)

    with report_file as report, benchmark_file as bdata:
        pipeline = [
            discoverer.get_iter,
        ]

        if options.dryrun:
            pipeline.append(dryrun)
        else:
            if options.pre_announce:
                options.num_procs = 1

            runner = ConcurrentTestRunner(options, queue)

            pipeline.append(runner.get_iter)

            if options.show_deprecations or options.deprecations_report:
                pipeline.append(DeprecationsReport(options).get_iter)

            if options.benchmark:
                pipeline.append(BenchmarkWriter(stream=bdata).get_iter)

            if options.durations:
                pipeline.append(DurationSummary(options).get_iter)

            if options.compact:
                verbose = -1
            else:
                verbose = int(options.verbose)

            pipeline.extend([
                ResultPrinter(options, verbose=verbose).get_iter,
                ResultSummary(options).get_iter,
            ])
            if not options.noreport:
                # print verbose results and summary to a report file
                if options.durations:
                    pipeline.append(DurationSummary(options, stream=report).get_iter)

                pipeline.extend([
                    ResultPrinter(options, report, verbose=1).get_iter,
                    ResultSummary(options, stream=report).get_iter,
                ])

        if options.maxtime > 0:
            pipeline.append(TimeFilter(options.maxtime).get_iter)

        if options.save_fails:
            pipeline.append(FailFilter().get_iter)

        retval = run_pipeline(tests, pipeline, options.disallow_skipped)

        finalize_coverage(options)

        if manager is not None:
            manager.shutdown()

    return retval


def run_tests(args=None):
    """This can be executed from within an "if __name__ == '__main__'" block
    to execute the tests found in that module.
    """
    if args is None:
        args = []
    sys.exit(main(list(args) + [__import__('__main__').__file__]))


if __name__ == '__main__':
    sys.exit(main())



import sys
import os
import site

from io import StringIO

class DeprecationsReport(object):
    """Generates a report of all Deprecation warnings raised during testing."""

    def __init__(self, options, stream=sys.stdout):
        self.stream = stream
        self.options = options
        self.startdir = os.getcwd()

    def get_iter(self, input_iter):

        deprecations = {}

        for test in input_iter:
            for msg, locs in test.deprecations.items():
                deprecations[msg] = deprecations.get(msg, set()) | locs
            yield test

        report = self.generate_report(deprecations)

        if self.options.show_deprecations:
            self.stream.write(report)

        if self.options.deprecations_report:
            with open(self.options.deprecations_report, 'w') as stream:
                stream.write(report)

    def generate_report(self, deprecations):
        report = StringIO()

        title = "Deprecations Report"
        eqs = "=" * len(title)

        write = report.write

        write("\n{}\n{}\n\n".format(title, eqs))

        count = len(deprecations)

        write("{} unique deprecation warnings were captured{}\n\n".format(count, ':' if count else '.'))

        if count == 0 and self.options.disallow_deprecations:
            write("\nDeprecation warnings have been raised as Exceptions\n"
                  "due to the use of the --disallow_deprecations option,\n"
                  "so no deprecation warnings have been captured.")

        startdir = self.startdir
        try:
            sitedirs = site.getsitepackages()
        except:
            # python older than 3.2
            sitedirs = []

        def trim_path(path):
            # trim the start directory or the site-packages directory from the path
            if path.startswith(startdir):
                path = path[len(startdir):]
            else:
                for d in sitedirs:
                    if path.startswith(d):
                        path = path[len(d):]
                        break
            return path

        for msg in sorted(deprecations):
            write("--\n{}\n\n".format(msg))
            for filename, lineno, test_spec in deprecations[msg]:
                write("    {}, line {}\n".format(trim_path(filename), lineno))
                if test_spec:
                    write("    [{}]\n\n".format(trim_path(test_spec)))

        if count > 0:
            write("\n\nFor a stack trace of reported deprecations, run the "
                  "identified test with the --disallow_deprecations option.")

        return report.getvalue()




import sys

from testflo.util import elapsed_str

_result_map = {
    ('FAIL', False): 'F',
    ('FAIL', True): 'X',  # expected failure
    ('SKIP', False): 'S',
    ('SKIP', True): 'S',
    ('OK', False): '.',
    ('OK', True): 'U',  # unexpected success
}

class ResultPrinter(object):
    """Prints the status and error message (if any) of each Test object
    after its test has been run if verbose is True.  If verbose is False,
    it displays a dot for each successful test, but skips or failures are
    still displayed in verbose form.
    """

    def __init__(self, options, stream=sys.stdout, verbose=0):
        self.stream = stream
        self.options = options
        self.verbose = verbose

    def get_iter(self, input_iter):
        for result in input_iter:
            self._print_result(result)
            yield result

    def _print_result(self, result):
        stream = self.stream

        stats = elapsed_str(result.elapsed())

        if ((result.expected_fail and result.status != 'FAIL') or
            (not result.expected_fail and result.status == 'FAIL')):
            show_msg = True
        else:
            show_msg = False

        if (self.verbose == 0 and (result.err_msg and show_msg)) or self.verbose > 0:
            if result.mpi and result.nprocs > 0:
                run_type = '(mpi) '
            elif result.isolated:
                run_type = '(isolated) '
            else:
                run_type = ''

            if result.err_msg:
                stream.write("%s%s ... %s (%s, %d MB)\n%s\n" % (
                                                     run_type,
                                                     result.spec,
                                                     result.status,
                                                     stats, result.memory_usage,
                                                     result.err_msg))
            else:
                stream.write("%s%s ... %s (%s, %d MB)\n" % (
                                                    run_type,
                                                    result.spec,
                                                    result.status,
                                                    stats, result.memory_usage))
        else:
            stream.write(_result_map[(result.status, result.expected_fail)])
            if self.options.pre_announce:
                stream.write('\n')

        stream.flush()




import sys
import time

from testflo.util import elapsed_str

class ResultSummary(object):
    """Writes a test summary after all tests are run."""

    def __init__(self, options, stream=sys.stdout):
        self.stream = stream
        self.options = options
        self._start_time = time.time()

    def get_test_name(self, test):
        if self.options.full_path:
            return test.spec
        else:
            return test.short_name()

    def get_iter(self, input_iter):
        oks = 0
        total = 0
        fails = []
        skips = []
        test_sum_time = 0.

        write = self.stream.write

        for test in input_iter:
            total += 1

            if test.status == 'OK':
                if test.expected_fail:
                    fails.append(self.get_test_name(test))
                else:
                    oks += 1
                test_sum_time += (test.end_time-test.start_time)
            elif test.status == 'FAIL':
                if test.expected_fail:
                    oks += 1
                else:
                    fails.append(self.get_test_name(test))
                test_sum_time += (test.end_time-test.start_time)
            elif test.status == 'SKIP':
                skips.append(self.get_test_name(test))

            yield test

        # now summarize the run
        if skips and (self.options.verbose or self.options.show_skipped):
            write("\n\nThe following tests were skipped:\n")
            for s in sorted(skips):
                write(s)
                write('\n')

        if fails:
            write("\n\nThe following tests failed:\n")
            for f in sorted(fails):
                write(f)
                write('\n')
        else:
            write("\n\nOK")

        write("\n\nPassed:  %d\nFailed:  %d\nSkipped: %d\n" %
                            (oks, len(fails), len(skips)))

        wallclock = time.time() - self._start_time

        s = "" if total == 1 else "s"
        if self.options.isolated:
            procstr = " in isolated processes"
        else:
            procstr = " using %d processes" % self.options.num_procs
        write("\n\nRan %d test%s%s\nWall clock time:   %s\n\n" %
                      (total, s, procstr, elapsed_str(wallclock)))



import sys

from testflo.util import _get_parser

_options = None

def get_options(args=None):
    global _options
    if _options is None:
        if args is None:
            args = sys.argv[1:]

        _options = _get_parser().parse_args(args)
    return _options



import os
import sys

import socket


# pickling the queue proxy gets rid of the authkey, so use a fixed authkey here
# for server and clients
_testflo_authkey = b'foobarxxxx'

def get_server_queue():
    from multiprocessing.managers import SyncManager
    #FIXME: some OSX users were getting "Can't assign requested address" errors
    # if we use socket.gethostname() for the address. Changing it to
    # 'localhost' seems to fix the issue, but I don't know why. We had to
    # use socket.gethostname() in order to get our benchmark tests to run
    # using qsub on a linux cluster, so with this 'fix', testflo benchmark tests
    # will likely not work on a cluster of OSX machines.
    if sys.platform == 'darwin':
        addr = 'localhost'
    else:
        addr = socket.gethostname()

    manager = SyncManager(address=(addr, 0), authkey=_testflo_authkey)
    manager.start()
    return manager, manager.Queue()

def get_client_queue():
    from multiprocessing.managers import RebuildProxy, AutoProxy, Token
    qstr = os.environ.get('TESTFLO_QUEUE')

    if qstr:
        # if TESTFLO_QUEUE is set, use that info to create a proxy that
        # points to the shared Queue.
        addr0, addr1, token_id = qstr.split(':')
        tok = Token('Queue', (addr0, int(addr1)), token_id)
        queue = RebuildProxy(AutoProxy, tok, 'pickle',
                             {'authkey':_testflo_authkey})
    else:
        queue = None

    return queue



from __future__ import print_function

import os
import sys
import time
import warnings
import traceback
from inspect import isclass
import subprocess
from contextlib import contextmanager, nullcontext

from types import FunctionType
from io import StringIO

from unittest import TestCase, SkipTest
from unittest.case import _UnexpectedSuccess

from testflo.cover import start_coverage, stop_coverage

from testflo.util import get_module, ismethod, get_memory_usage, \
                         get_testpath, _options2args
from testflo.devnull import DevNull


from distutils import spawn
mpirun_exe = None
if spawn.find_executable("mpirun") is not None:
    mpirun_exe = "mpirun"
elif spawn.find_executable("mpiexec") is not None:
    mpirun_exe = "mpiexec"


def add_queue_to_env(queue):
    """Store enough info in the env to be able to create a proxy to
    the queue in a subprocess.
    """
    addr = queue._token.address
    os.environ['TESTFLO_QUEUE'] = "%s:%s:%s" % (addr[0], addr[1],
                                                queue._token.id)


class FakeComm(object):
    def __init__(self):
        self.rank = 0
        self.size = 1


# create a copy of sys.path with an extra entry at the beginning so that
# we can quickly replace the first entry with the curent test's dir rather
# than constantly copying the whole sys.path
_testing_path = ['.'] + sys.path


@contextmanager
def testcontext(test):
    global _testing_path
    old_sys_path = sys.path

    _testing_path[0] = test.test_dir
    sys.path = _testing_path

    try:
        yield
    except Exception:
        test.status = 'FAIL'
        test.err_msg = traceback.format_exc()
    finally:
        sys.path = old_sys_path


class Test(object):
    """Contains the path to the test function/method, status
    of the test (if finished), error and stdout messages (if any),
    start/end times and resource usage data.
    """

    def __init__(self, testspec, options):
        self.spec = testspec
        self.options = options

        testpath, rest = get_testpath(testspec)
        self.test_dir = os.path.dirname(testpath)

        self.status = None
        self.err_msg = ''
        self.mpi = False

        self.memory_usage = 0
        self.nprocs = 0
        self.isolated = False
        self.start_time = 0
        self.end_time = 0
        self.modpath = None
        self.tcasename = None
        self.funcname = None
        self.load = (0.0, 0.0, 0.0)
        self.expected_fail = False
        self._mod_fixture_first = False
        self._mod_fixture_last = False
        self._tcase_fixture_first = False
        self._tcase_fixture_last = False

        self.deprecations = {}

        self._get_test_info()

    def __iter__(self):
        """Allows Test to be iterated over so we don't have to check later
        for Test vs. iter of Tests.
        """
        return iter((self,))

    def _get_test_info(self):
        """Get the test's module, testcase (if any), function name,
        N_PROCS (for mpi tests) and ISOLATED and set our attributes.
        """
        with testcontext(self):
            try:
                mod, self.tcasename, self.funcname = _parse_test_path(self.spec)
                self.modpath = mod.__name__
            except Exception:
                self.status = 'FAIL'
                self.err_msg = traceback.format_exc()
            else:
                if self.funcname is None:
                    self.status = 'FAIL'
                    self.err_msg = 'ERROR: test function not specified.'
                else:
                    if self.tcasename is not None:
                        testcase = getattr(mod, self.tcasename)
                        self.nprocs = getattr(testcase, 'N_PROCS', 0)
                        self.isolated = getattr(testcase, 'ISOLATED', False)

        if self.err_msg:
            self.start_time = self.end_time = time.time()

    def _run_sub(self, cmd, queue, env):
        """
        Run a command in a subprocess.
        """
        try:
            add_queue_to_env(queue)

            if self.options.nocapture:
                stdout = subprocess.PIPE
                stderr = subprocess.STDOUT
            else:
                stdout = subprocess.DEVNULL
                stderr = subprocess.PIPE

            p = subprocess.run(cmd, stdout=stdout, stderr=stderr, env=env,
                               timeout=self.options.timeout, universal_newlines=True)

            if p.returncode != 0:
                self.status = 'FAIL'
                self.err_msg = p.stdout if self.options.nocapture else p.stderr
                result = self
            else:
                if self.options.nocapture:
                    print(p.stdout)
                result = queue.get()
        except:
            # we generally shouldn't get here, but just in case,
            # handle it so that the main process doesn't hang at the
            # end when it tries to join all of the concurrent processes.
            self.status = 'FAIL'
            self.err_msg = traceback.format_exc()
            result = self

        return result

    def _run_isolated(self, queue):
        """This runs the test in a subprocess,
        then returns the Test object.
        """

        cmd = [sys.executable,
               os.path.join(os.path.dirname(__file__), 'isolatedrun.py'),
               self.spec]

        try:
            result = self._run_sub(cmd, queue, os.environ)
        except:
            # we generally shouldn't get here, but just in case,
            # handle it so that the main process doesn't hang at the
            # end when it tries to join all of the concurrent processes.
            self.status = 'FAIL'
            self.err_msg = traceback.format_exc()
            result = self

        result.isolated = True

        return result

    def _run_mpi(self, queue):
        """This runs the test using mpirun in a subprocess,
        then returns the Test object.
        """
        try:
            if mpirun_exe is None:
                raise Exception("mpirun or mpiexec was not found in the system path.")

            cmd =  [mpirun_exe, '-n', str(self.nprocs),
                   sys.executable,
                   os.path.join(os.path.dirname(__file__), 'mpirun.py'),
                   self.spec] + _options2args()

            result = self._run_sub(cmd, queue, os.environ)

        except:
            # we generally shouldn't get here, but just in case,
            # handle it so that the main process doesn't hang at the
            # end when it tries to join all of the concurrent processes.
            self.status = 'FAIL'
            self.err_msg = traceback.format_exc()
            result = self

        result.mpi = True

        return result

    def run(self, queue=None):
        """Runs the test, assuming status is not already known."""
        if self.status is not None:
            # premature failure occurred (or dry run), just return
            return self

        MPI = None
        if self.nprocs > 0 and not self.options.nompi:
            try:
                from mpi4py import MPI
            except ImportError:
                pass
            else:
                if queue is not None:
                    return self._run_mpi(queue)
        elif self.options.isolated:
            return self._run_isolated(queue)

        with testcontext(self):
            testpath, _ = get_testpath(self.spec)
            _, mod = get_module(testpath)

            testcase = getattr(mod, self.tcasename) if self.tcasename is not None else None
            funcname, nprocs = (self.funcname, self.nprocs)

            mod_setup = mod_teardown = tcase_setup = tcase_teardown = None

            if self._mod_fixture_first:
                mod_setup = getattr(mod, 'setUpModule', None)
            if self._mod_fixture_last:
                mod_teardown = getattr(mod, 'tearDownModule', None)

            if testcase is not None:
                if self._tcase_fixture_first:
                    tcase_setup = getattr(testcase, 'setUpClass', None)
                if self._tcase_fixture_last:
                    tcase_teardown = getattr(testcase, 'tearDownClass', None)

                parent = testcase(methodName=funcname)
                # if we get here and nprocs > 0, we need
                # to set .comm in our TestCase instance.
                if nprocs > 0:
                    if MPI is not None and not self.options.nompi:
                        parent.comm = MPI.COMM_WORLD
                    else:
                        parent.comm = FakeComm()

                setup = getattr(parent, 'setUp', None)
                teardown = getattr(parent, 'tearDown', None)
            else:
                parent = mod
                setup = teardown = None

            if self.options.nocapture:
                outstream = sys.stdout
            else:
                outstream = DevNull()
            errstream = StringIO()

            done = False
            expected = expected2 = expected3 = False

            try:
                old_err = sys.stderr
                old_out = sys.stdout
                sys.stdout = outstream
                sys.stderr = errstream

                start_coverage()

                self.start_time = time.time()

                catch_deps = self.options.show_deprecations or self.options.deprecations_report
                raise_deps = self.options.disallow_deprecations

                with warnings.catch_warnings(record=True) if catch_deps else nullcontext() as w:

                    if raise_deps:
                        # raise deprecation warnings as exceptions
                        warnings.filterwarnings("error", category=DeprecationWarning)
                        warnings.filterwarnings("error", category=PendingDeprecationWarning)
                    elif catch_deps:
                        # catch all locations where deprecation warning is triggered
                        warnings.filterwarnings("always", category=DeprecationWarning)
                        warnings.filterwarnings("always", category=PendingDeprecationWarning)

                    # if there's a module setup, run it
                    if mod_setup:
                        status, expected = _try_call(mod_setup)
                        if status != 'OK':
                            done = True
                            mod_teardown = None # don't do teardown if setup failed

                    # handle @unittest.skip class decorator
                    if not done and hasattr(parent, '__unittest_skip__') and parent.__unittest_skip__:
                        sys.stderr.write("%s\n" % parent.__unittest_skip_why__)
                        status = 'SKIP'
                        done = True
                        tcase_setup = None
                        tcase_teardown = None

                    if tcase_setup:
                        status, expected = _try_call(tcase_setup)
                        if status != 'OK':
                            done = True
                            tcase_teardown = None

                    # if there's a setUp method, run it
                    if not done and setup:
                        status, expected = _try_call(setup)
                        if status != 'OK':
                            done = True

                    if not done:
                        status, expected2 = _try_call(getattr(parent, funcname))

                    if not done and teardown:
                        tdstatus, expected3 = _try_call(teardown)
                        if status == 'OK':
                            status = tdstatus

                    if tcase_teardown:
                        _try_call(tcase_teardown)

                    if mod_teardown:
                        _try_call(mod_teardown)

                    self.end_time = time.time()
                    self.status = status
                    self.err_msg = errstream.getvalue()
                    self.memory_usage = get_memory_usage()
                    self.expected_fail = expected or expected2 or expected3

                    if sys.platform == 'win32':
                        self.load = (0.0, 0.0, 0.0)
                    else:
                        self.load = os.getloadavg()

                    if catch_deps and w:
                        deprecations = [wm for wm in w
                                        if wm.category is DeprecationWarning
                                        or wm.category is PendingDeprecationWarning]
                        for wm in deprecations:
                            msg = str(wm.message)
                            dep = self.deprecations.get(msg, set())
                            dep.add((wm.filename, wm.lineno, self.spec))
                            self.deprecations[msg] = dep

            finally:
                stop_coverage()

                sys.stderr = old_err
                sys.stdout = old_out

        return self

    def elapsed(self):
        return self.end_time - self.start_time

    def short_name(self):
        """Returns the testspec with only the file's basename instead
        of its full path.
        """
        testpath, rest = get_testpath(self.spec)
        fname = os.path.basename(testpath)
        return ':'.join((fname, rest))

    def __str__(self):
        if self.err_msg:
            return "%s: %s\n%s" % (self.spec, self.status, self.err_msg)
        else:
            return "%s: %s" % (self.spec, self.status)


def _parse_test_path(testspec):
    """Return a tuple of the form (module, testcase, func)
    based on the given testspec.

    The format of testspec is one of the following:
        <module>
        <module>:<testcase>
        <module>:<testcase>.<method>
        <module>:<function>

    where <module> is either the python module path or the actual
    file system path to the .py file.  A value of None in the tuple
    indicates that that part of the testspec was not present.
    """
    testpath, rest = get_testpath(testspec)
    _, mod = get_module(testpath)

    funcname = tcasename = None

    if rest:
        objname, _, funcname = rest.partition('.')
        obj = getattr(mod, objname)
        if isclass(obj) and issubclass(obj, TestCase):
            tcasename = objname
            if funcname:
                meth = getattr(obj, funcname)
                if not ismethod(meth):
                    raise TypeError("'%s' is not a method." % rest)
        elif isinstance(obj, FunctionType):
            funcname = obj.__name__
        else:
            raise TypeError("'%s' is not a TestCase or a function." %
                            objname)

    return (mod, tcasename, funcname)


def _try_call(func):
    """Calls the given method, captures stdout and stderr,
    and returns the status (OK, SKIP, FAIL).
    """
    status = 'OK'
    if getattr(func, '__unittest_expecting_failure__', False):
        expected = True
    else:
        expected = False
    try:
        func()
    except SkipTest as e:
        status = 'SKIP'
        sys.stderr.write(str(e))
    except _UnexpectedSuccess:
        status = 'OK'
        expected = True
    except Exception:
        status = 'FAIL'
        sys.stderr.write(traceback.format_exc())

    return status, expected




class DevNull(object):
    """A class for throwaway stream output."""

    def write(self, s):
        pass

    def writelines(self, iterable):
        pass

    def flush(self):
        pass

    def isatty(self):
        return False




"""
This is meant to be executed as a subprocess of testflo.

"""

if __name__ == '__main__':
    import sys
    import os
    import traceback

    from testflo.test import Test
    from testflo.cover import save_coverage
    from testflo.qman import get_client_queue
    from testflo.options import get_options

    queue = get_client_queue()
    os.environ['TESTFLO_QUEUE'] = ''

    options = get_options()

    try:
        try:
            test = Test(sys.argv[1], options)
            test.nocapture = True # so we don't lose stdout
            test.run()
        except:
            print(traceback.format_exc())
            test.status = 'FAIL'
            test.err_msg = traceback.format_exc()

        save_coverage()

    except:
        test.err_msg = traceback.format_exc()
        test.status = 'FAIL'

    sys.stdout.flush()
    sys.stderr.flush()

    queue.put(test)



from setuptools import setup

import re

__version__ = re.findall(
    r"""__version__ = ["']+([0-9\.\-dev]*)["']+""",
    open('testflo/__init__.py').read(),
)[0]

setup(name='testflo',
      version=__version__,
      description="A simple flow-based testing framework",
      long_description="""
        usage: testflo [options]

        positional arguments:
          test                  A test method, test case, module, or directory to run.

        optional arguments:
          -h, --help            show this help message and exit
          -c FILE, --config FILE
                                Path of config file where preferences are specified.
          -t FILE, --testfile FILE
                                Path to a file containing one testspec per line.
          --maxtime TIME_LIMIT  Specifies a time limit in seconds for tests to be
                                saved to the quicktests.in file.
          -n NUM_PROCS, --numprocs NUM_PROCS
                                Number of processes to run. By default, this will use
                                the number of CPUs available. To force serial
                                execution, specify a value of 1.
          -o FILE, --outfile FILE
                                Name of test report file. Default is
                                testflo_report.out.
          -v, --verbose         Include testspec and elapsed time in screen output.
                                Also shows all stderr output, even if test doesn't
                                fail
          --compact             Limit output to a single character for each test.
          --dryrun              Don't actually run tests, but print which tests would
                                have been run.
          --pre_announce        Announce the name of each test before it runs. This
                                can help track down a hanging test. This automatically
                                sets -n 1.
          -f, --fail            Save failed tests to failtests.in file.
          --full_path           Display full test specs instead of shortened names.
          -i, --isolated        Run each test in a separate subprocess.
          --nompi               Force all tests to run without MPI. This can be useful
                                for debugging.
          -x, --stop            Stop after the first test failure, or as soon as
                                possible when running concurrent tests.
          -s, --nocapture       Standard output (stdout) will not be captured and will
                                be written to the screen immediately.
          --coverage            Perform coverage analysis and display results on
                                stdout
          --coverage-html       Perform coverage analysis and display results in
                                browser
          --coverpkg PKG        Add the given package to the coverage list. You can
                                use this option multiple times to cover multiple
                                packages.
          --cover-omit FILE     Add a file name pattern to remove it from coverage.
          -b, --benchmark       Specifies that benchmarks are to be run rather than
                                tests, so only files starting with "benchmark\_" will
                                be executed.
          -d FILE, --datafile FILE
                                Name of benchmark data file. Default is
                                benchmark_data.csv.
          --noreport            Don't create a test results file.
          -m GLOB, --match GLOB, --testmatch GLOB
                                Pattern to use for test discovery. Multiple patterns
                                are allowed.
          --timeout TIMEOUT     Timeout in seconds. Test will be terminated if it
                                takes longer than timeout. Only works for tests
                                running in a subprocess (MPI and isolated).
      """,
      classifiers=[
          'Development Status :: 4 - Beta',
          'License :: OSI Approved :: Apache Software License',
          'Natural Language :: English',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: POSIX :: Linux',
          'Operating System :: Microsoft :: Windows',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: Implementation :: CPython',
      ],
      license='Apache 2.0',
      install_requires=[
        'coverage<5.0'
      ],
      packages=['testflo'],
      entry_points="""
          [console_scripts]
          testflo=testflo.main:main
      """
      )



import sys
import time

class BenchmarkWriter(object):
    """Writes benchmark data to a file for postprocessing.
       Data is written as comma separated values (CSV)
    """

    def __init__(self, stream=sys.stdout):
        self.timestamp = time.time()
        self.stream = stream

    def get_iter(self, input_iter):
        for result in input_iter:
            self._write_data(result)
            yield result

    def _write_data(self, result):
        stream = self.stream
        stream.write('%d,%s,%s,%f,%f,%f,%f,%f\n' % (
            self.timestamp,
            result.spec,
            result.status,
            result.elapsed(),
            result.memory_usage,
            result.load[0],
            result.load[1],
            result.load[2]
        ))
        stream.flush()




import traceback
from inspect import getmembers, isclass, isfunction
from importlib import import_module
from unittest import TestCase

from os.path import basename, dirname, isdir

from testflo.util import find_files, get_module, get_testpath, ismethod
from testflo.test import Test

def _has_class_fixture(tcase):
    if tcase is not None:
        for klass in tcase.__mro__:
            if klass is TestCase:
                break
            if 'setUpClass' in klass.__dict__ or 'tearDownClass' in klass.__dict__:
                return  True
    return False


class TestDiscoverer(object):

    def __init__(self, options, module_pattern='test*.py',
                       func_match=lambda f: fnmatchcase(f, 'test*'),
                       dir_exclude=None):
        self.options = options
        self.module_pattern = module_pattern
        self.func_match = func_match
        self.dir_exclude = dir_exclude

        # to support module and class fixtures, we need to be able to
        # process all tests in a module or TestCase in the same process,
        # so these are to keep track of which tests need to be grouped
        # together.
        self._mod_fixture_groups = {}
        self._tcase_fixture_groups = {}

    def get_iter(self, input_iter):
        """Returns an iterator of Test objects
        based on the starting list of directories/modules/testspecs.
        """
        seen = set()
        for tests in input_iter:
            if isdir(tests):
                itr = self._dir_iter
            else:
                itr = self._testspec_iter

            for result in itr(tests):
                if result.spec not in seen:
                    seen.add(result.spec)
                    result = self._filter(result)
                    if result is not None:
                        yield result

        # Every test left has been group together either by module or
        # TestCase or both, due to the presence of module or testcase class level
        # setup/teardown, and we need to run each group on the same
        # process so that we can execute the module or class level setup/teardown
        # only once while impacting all of the tests in that group.
        new_tcase_groups = []
        for tcase, tests in self._tcase_fixture_groups.items():
            tests = sorted(tests, key=lambda t: t.spec)

            # mark the first and last tests so that we know when to
            # run setUpClass and tearDownClass
            tests[0]._tcase_fixture_first = True
            tests[-1]._tcase_fixture_last = True

            # check to see if this TestCase is part of a module with setUpModule/tearDownModule
            if tests[0].modpath in self._mod_fixture_groups:
                # these tests are already part of a module fixture, so we
                # don't want to execute them a second time
                continue

            new_tcase_groups.append(tests)

        # yield any tests that are grouped because of a module level fixture.
        for tests in self._mod_fixture_groups.values():
            tests = sorted(tests, key=lambda t: t.spec)

            # mark the first and last tests so that we know when to
            # run setUpModule and tearDownModule
            tests[0]._mod_fixture_first = True
            tests[-1]._mod_fixture_last = True
            yield tests  # yield them together as a group

        # yield grouped tests for all remaining TestCases with setUpClass/tearDownClass
        for tests in new_tcase_groups:
            yield tests

    def _filter(self, test):
        """
        If the given test is part of a module with setUpModule/tearDownModule
        and/or part of a TestCase with setUpClass/tearDownClass, then save it
        for later, else return it.
        """
        if test.status is not None:
            return test

        mod = import_module(test.modpath)
        if test.modpath in self._mod_fixture_groups:
            self._mod_fixture_groups[test.modpath].append(test)
        elif hasattr(mod, 'setUpModule') or hasattr(mod, 'tearDownModule'):
            self._mod_fixture_groups[test.modpath] = [test]

        full_tcase = (test.modpath, test.tcasename)
        testcase = getattr(mod, test.tcasename) if test.tcasename else None
        if full_tcase in self._tcase_fixture_groups:
            self._tcase_fixture_groups[full_tcase].append(test)
        elif _has_class_fixture(testcase):
            self._tcase_fixture_groups[full_tcase] = [test]

        if not (test.modpath in self._mod_fixture_groups or
                full_tcase in self._tcase_fixture_groups):
            return test

    def _dir_iter(self, dname):
        """Iterate over all tests in modules found in the given
        directory and its subdirectories. Returns an iterator
        of Test objects.
        """
        for f in find_files(dname, match=self.module_pattern,
                                   direxclude=self.dir_exclude):
            if not basename(f).startswith('__init__.'):
                for result in self._module_iter(f):
                    yield result

    def _module_iter(self, filename):
        """Returns an iterator of Test objects for the contents of
        the given python module file.
        """

        try:
            fname, mod = get_module(filename)
        except:
            t =  Test(filename, self.options)
            t.status = 'FAIL'
            t.err_msg=traceback.format_exc()
            yield t
        else:
            if basename(fname).startswith('__init__.'):
                for result in self._dir_iter(dirname(fname)):
                    yield result
            else:
                for name, obj in getmembers(mod):
                    if isclass(obj) and issubclass(obj, TestCase):
                        for result in self._testcase_iter(filename, obj):
                            yield result

                    elif isfunction(obj) and self.func_match(name):
                        yield Test(':'.join((filename, obj.__name__)), self.options)

    def _testcase_iter(self, fname, testcase):
        """Returns an iterator of Test objects coming from a given
        TestCase class.
        """
        tcname = ':'.join((fname, testcase.__name__))
        for name, method in getmembers(testcase, ismethod):
            if self.func_match(name):
                yield Test('.'.join((tcname, name)), self.options)

    def _testspec_iter(self, testspec):
        """Returns an iterator of Test objects found in the
        module/testcase/method specified in testspec.  The format of
        testspec is one of the following:
            <module>
            <module>:<testcase>
            <module>:<testcase>.<method>
            <module>:<function>

        where <module> is either the python module path or the actual
        file system path to the .py file.
        """

        module, rest = get_testpath(testspec)
        if rest:
            tcasename, _, method = rest.partition('.')
            if method:
                yield Test(testspec, self.options)
            else:  # could be a test function or a TestCase
                try:
                    fname, mod = get_module(module)
                except:
                    t = Test(testspec, self.options)
                    t.status = 'FAIL'
                    t.err_msg = traceback.format_exc()
                    return
                try:
                    tcase = get_testcase(fname, mod, tcasename)
                except (AttributeError, TypeError):
                    t = Test(testspec, self.options)
                    t.status = 'FAIL'
                    t.err_msg = traceback.format_exc()
                    yield t
                else:
                    for test in self._testcase_iter(fname, tcase):
                        yield test
        else:
            for test in self._module_iter(module):
                yield test


def get_testcase(filename, mod, tcasename):
    """Given a module and the name of a TestCase
    class, return a TestCase class object or raise an exception.
    """

    try:
        tcase = getattr(mod, tcasename)
    except AttributeError:
        raise AttributeError("Couldn't find TestCase '%s' in module '%s'" %
                               (tcasename, filename))
    if issubclass(tcase, TestCase):
        return tcase
    else:
        raise TypeError("'%s' in file '%s' is not a TestCase." %
                        (tcasename, filename))
